# from zope
import Globals
import Acquisition
from Globals import Persistent, DTMLFile
import AccessControl
from AccessControl import ClassSecurityInfo
import OFS

# python modules
from urllib import quote
import string
from StringIO import StringIO

# Formulator
# for forms initialized at module level (needs form._realize_fields())
from Products.Formulator.DummyField import fields
# forms in normal code use fiels from this
from Products.Formulator import StandardFields
from Products.Formulator.Form import BasicForm, FormValidationError

# for XPath evaluation (from 4suite)
from xml import xpath
from xml.xpath.Context import Context

# for printing
from Products.ParsedXML import ExtraDOM

class ArgumentError(Exception):
    pass

class XPathMethod:
    security = ClassSecurityInfo()
   
    def __init__(self):
        # expected arguments to method
        self.arguments = []
        # expression
        self.expression = ""

    security.declareProtected('Use XPath Methods', 'evaluate')
    def evaluate(self, node, **args):
        """Call this XPath method on a node.
        Use keyword arguments to supply arguments.
        """
        for argument_name in self.arguments:
            if not args.has_key(argument_name):
                raise ArgumentError, ("Keyword argument '%s' expected "
                                      "but not found" % argument_name)
        if len(self.arguments) != len(args):
            raise ArgumentError, ("%s keyword arguments expected "
                                  "but found %s" % (len(self.arguments),
                                                    len(args)))

        # I'm not sure why $variables are in namespaces, so hack around
        # it for now giving '' namespace URI
        argdict = {}
        for key, value in args.items():
            argdict[('', key)] = str(value)
            
        if hasattr(node, 'getDOMObj'):
            rewrap = 1
            wrapDOMObj = node.wrapDOMObj
            node = node.getDOMObj()
        else:
            rewrap = 0
        context = Context(node, 0, 0)
        context.setVarBindings(argdict)
        nodes = xpath.Evaluate(self.expression, context=context)
        if rewrap:
            return map(wrapDOMObj, nodes)
        else:
            return nodes

    security.declareProtected('Use XPath Methods', 'evaluate_multiple')
    def evaluate_multiple(self, nodes, **args):
        """Search in a number of nodes (could be in separate
        documents."""
        result = []
        for node in nodes:
            result.extend(apply(self.evaluate, [node], args))
        # FIXME: perhaps erase possible duplicates?
        return result
    
Globals.InitializeClass(XPathMethod)

def create_edit_form():
    form = BasicForm("manage_edit", "POST")
    form.add_field(fields.StringField(
        "title",
        title="Title",
        required=0,
        display_width=30))
    form.add_field(fields.LinesField(
        "arguments",
        title="Arguments",
        required=0,
        width=30,
        height=2))
    form.add_field(fields.StringField(
        "expression",
        title="Expression",
        required=1,
        display_width=30))
    
    return form

def create_test_form(arguments=None):
    arguments = arguments or []
    form = BasicForm("manage_test", "POST")
    form.add_field(StandardFields.StringField(
        "doc_id",
        title="Parsed XML Document id",
        required=1,
        display_width=20))
    # FIXME: what if there is an argument called doc_id?
    for argument in arguments:
        form.add_field(StandardFields.StringField(
            argument,
            title=argument,
            required=1,
            display_width=20))
    return form

class ZMIXPathMethod(
    Acquisition.Implicit,
    Persistent,
    OFS.SimpleItem.Item,
    XPathMethod):

    security = ClassSecurityInfo()

    security.declareObjectProtected('View')

    meta_type = "XPath Method"
    
    manage_options = (
        (
        {'label': 'Edit', 'action':'manage_main',
         'help':('XPathMethods', 'methodEdit.txt')},
        {'label': 'Test', 'action':'manage_testForm',
         'help':('XPathMethods', 'methodTest.txt')},
        ) +
        AccessControl.Role.RoleManager.manage_options +
        OFS.SimpleItem.Item.manage_options
        )
        
    def __init__(self, id, title):
        """Initialize form.
        id -- id of form
        title -- the title of the form
        """
        ZMIXPathMethod.inheritedAttribute('__init__')(self)
        self.id = id
        self.title = title
        self.test_form = create_test_form()
    
    security.declareProtected('View management screens', 'manage_main')
    manage_main = DTMLFile('dtml/methodEdit', globals())

    security.declareProtected('View management screens', 'edit_form')
    edit_form = create_edit_form()
    
    security.declareProtected('Change XPath Methods', 'manage_edit')
    def manage_edit(self, REQUEST):
        """Submit edit form.
        """
        try:
            result = self.edit_form.validate_all(REQUEST)
        except FormValidationError, e:
            message = get_manage_tabs_message(e)            
            return self.manage_main(self, REQUEST,
                                    manage_tabs_message=message)

        self.title = result['title']
        self.arguments = result['arguments']
        self.test_form = create_test_form(self.arguments)
        self.expression = result['expression']
        return self.manage_main(self, REQUEST,
                                manage_tabs_message="Settings changed.")
        
    security.declareProtected('View management screens', 'manage_testForm')
    manage_testForm = DTMLFile('dtml/methodTest', globals())

    security.declareProtected('View management screens', 'test_form')
    
    security.declareProtected('View management screens', 'manage_test')
    def manage_test(self, REQUEST):
        """Test this XPath method.
        """
        try:
            result = self.test_form.validate_all(REQUEST)
        except FormValidationError, e:
            message = get_manage_tabs_message(e)            
            return self.manage_testForm(self, REQUEST,
                                        manage_tabs_message=message)
        doc_id = result['doc_id']

        # constructing dictionary so we can use apply with evaluate();
        # a bit redundant but we want evaluate() to take keyword arguments
        # when called directly
        argdict = {}
        for argument in self.arguments:
            argdict[argument] = result[argument]
            
        nodes = apply(self.evaluate,
                      (getattr(self, doc_id).documentElement.getDOMObj(),),
                      argdict)
        result = []
        for node in nodes:
            result.append(ExtraDOM.writeStream(node).getvalue())
        return self.manage_testForm(self, REQUEST,
                                    result=result)
        
Globals.InitializeClass(ZMIXPathMethod)    

manage_addForm = DTMLFile('dtml/methodAdd', globals())

def manage_add(self, id, title="", REQUEST=None):
    """Add XPath Method to folder.
    """
    id = self._setObject(id, ZMIXPathMethod(id, title))
    add_and_edit(self, id, REQUEST)
    return ''

def add_and_edit(self, id, REQUEST):
    """Helper method to point to the object's management screen if
    'Add and Edit' button is pressed.
    id -- id of the object we just added
    """
    if REQUEST is None:
        return
    try:
        u = self.DestinationURL()
    except:
        u = REQUEST['URL1']
    if REQUEST['submit'] == " Add and Edit ":
        u = "%s/%s" % (u, quote(id))
    REQUEST.RESPONSE.redirect(u+'/manage_main')

def get_manage_tabs_message(e):
    # create a failure message containing all errors
    messages = ["Validation error(s)"]
    # loop through all error texts and generate messages for it
    for error in e.errors:
        messages.append("%s: %s" % (error.field.get_value('title'),
                                    error.error_text))
    # join them all together in a big message
    return string.join(messages, "<br>")
